"""Typing hints"""
from typing import Optional, Any, Protocol

OpenInstance = Optional[Any]


class ComparableType(Protocol):
    def __lt__(self, other: Any) -> bool:
        ...

    def __gt__(self, other: Any) -> bool:
        ...



"""Testing the util.py module"""

from stream2py.util import contextualize_with_instance


class StreamHasNotBeenStarted(RuntimeError):
    """Raised when an action requires the stream to be 'on'"""


class Streamer:
    def __init__(self, iterable):
        self.iterable = iterable
        self.is_running = False
        self._read = None

    def enter(self):
        #         print(f'{type(self).__name__}.enter')
        self._read = iter(self.iterable).__next__
        self.is_running = True

    def exit(self, *exc):
        #         print(f'{type(self).__name__}.exit')
        self._read = None
        self.is_running = False

    __enter__, __exit__ = enter, exit

    def read(self):
        if not self.is_running:
            raise StreamHasNotBeenStarted(
                'The stream needs to be on/started (in a context) for that!'
            )
        return self._read()


def test_test_objects():
    """This test tests the test objects themselves, to make sure they have the
    expected behavior, commenting at the same time on what kind of behavior we'd
    like to have the power to create, which we get through contextualize_with_instance
    """
    # Testing the test objects

    s = Streamer('stream')

    # demo
    with s:
        assert s.read() == 's'
        assert s.read() == 't'

    # a reader test function
    def test_reader(reader):
        assert ''.join(reader() for _ in range(6)) == 'stream'

    # Normal case: With a context

    reader = s.read

    with s:
        test_reader(reader)

    # Normal case: Manual entering/exiting

    s.enter()
    reader = s.read
    test_reader(reader)
    s.exit()

    # But if we don't turn things on...

    reader = s.read
    try:
        # oops, forgot the enter s context
        test_reader(reader)
        it_worked = True
    except StreamHasNotBeenStarted as e:
        it_worked = False
        assert isinstance(e, StreamHasNotBeenStarted)
        assert e.args[0] == 'The stream needs to be on/started (in a context) for that!'
    assert not it_worked

    # But we can't turn the read method on -- it's not a context (it's instance is!)

    reader = s.read

    try:
        with reader:  # can we actually do this (answer: no! We can enter s, not s.read)
            test_reader(reader)
        it_worked = True
    except Exception as e:
        it_worked = False
        assert isinstance(e, AttributeError)
        assert e.args[0] == '__enter__'  # well yeah, reader doesn't have an __enter__!

    assert not it_worked

    # But with contextualize_with_instance, you can
    # (see test_contextualize_with_instance)


def test_contextualize_with_instance():
    """To understand the context of this test, the reader should consider the
    test_test_objects, or even better, see:

    issue: https://github.com/i2mint/stream2py/issues/18

    wiki: https://github.com/i2mint/stream2py/wiki/Forwarding-context-management

    """
    s = Streamer('stream')

    # a reader test function
    def test_reader(reader):
        assert ''.join(reader() for _ in range(6)) == 'stream'

    reader = contextualize_with_instance(s.read)

    try:
        with reader:  # now we can enter the reader!
            test_reader(reader)
        it_worked = True
    except Exception as e:
        it_worked = False

    assert it_worked  # Hurray!



"""
A BufferReader gives data access to any number of consumers and provides methods to seek data such
as next(), range(), head(), tail().  Each BufferReader instance has it's own cursor keeping track of
what data was last seen."""
__all__ = ['BufferReader']

from contextlib import suppress
import threading
import time
from functools import wraps, partialmethod, partial
from typing import Union

from stream2py.utility.locked_sorted_deque import RWLockSortedDeque


def defaulted_values(source_dict, defaults):
    """Returns a the values of source_dict where None values are replaced with the
    corresponding `default_dict` values.

    >>> defaults = {'a': 1, 'b': 2, 'c': 3}
    >>> tuple(defaulted_values({'a': None, 'b': 22, 'c': None}, defaults))
    (1, 22, 3)
    >>> tuple(defaulted_values({'a': False, 'b': True, 'c': None}, defaults))
    (False, True, 3)
    """
    return {
        k: v if v is not None else defaults[k] for k, v in source_dict.items()
    }.values()


class BufferReader:
    """Reader that is constructed from StreamBuffer.mk_reader()

    >>> from stream2py import StreamBuffer, BufferReader
    >>> from stream2py.examples.source_reader import SimpleCounterString
    >>>
    >>> source_reader = SimpleCounterString(start=0, stop=100)
    >>> stream_buffer = StreamBuffer(source_reader=source_reader, maxlen=100)
    >>> stream_buffer.start()
    >>>
    >>> buffer_reader = stream_buffer.mk_reader()
    >>> isinstance(buffer_reader, BufferReader)
    True
    >>> buffer_reader.is_stopped
    False
    >>> buffer_reader.last_item == None  # None until value is seen
    True
    >>> buffer_reader.next() # get first item
    's0'
    >>> buffer_reader.last_item # check last_item cursor
    's0'
    >>> # next will not update last_item cursor
    >>> buffer_reader.next(peek=True), buffer_reader.last_item
    ('s1', 's0')
    >>> # same as what was peeked but now cursor is updated
    >>> buffer_reader.next(), buffer_reader.last_item
    ('s1', 's1')
    >>> buffer_reader.range(start=5, stop=10)
    ['s5', 's6', 's7', 's8', 's9', 's10']
    >>> buffer_reader.last_item
    's10'
    >>> buffer_reader.range(start=5, stop=10, step=3), buffer_reader.last_item
    (['s5', 's8'], 's8')
    >>> buffer_reader.range(start=10, stop=15, peek=True)
    ['s10', 's11', 's12', 's13', 's14', 's15']
    >>> buffer_reader.last_item # unchanged with peek
    's8'
    >>> buffer_reader.last_key # key of last_item
    8
    >>> # only 9 is greater than last item key
    >>> buffer_reader.range(start=0, stop=9, only_new_items=True)
    ['s9']
    >>> buffer_reader.head(), buffer_reader.last_item, buffer_reader.last_key
    ('s0', 's0', 0)
    >>> buffer_reader.tail(), buffer_reader.last_item, buffer_reader.last_key
    ('s99', 's99', 99)
    >>>
    >>> stream_buffer.stop()
    >>>
    >>> buffer_reader.is_stopped
    True
    """

    read = None  # will be set by init

    def __init__(
        self,
        buffer: RWLockSortedDeque,
        source_reader_info: dict,
        stop_event: threading.Event,
        *,
        read_size=1,
        peek=False,
        strict_n=False,
        ignore_no_item_found=False,  # TODO: should this be True to be aligned with iter?
    ):
        """

        The `read_size`, `peek`, `strict_n` and `ignore_no_item_found` parameters
        are used as defaults of the
        `read` method. More importanty -- since the default read is used by
        `__next__` and therefore by `__iter__`, the `read_size` and `peek` these
        parameters define the behavior of

        :param buffer:
        :param source_reader_info:
        :param stop_event: threading.Event for source read loop
        :param read_size: number of items to return by default when reading.
            Is also used as the "chunk size" when iterating.
        :param peek: if True, by default, last_item cursor will not be updated during a
            read
        :param ignore_no_item_found: if True, by default return None when no next item
            instead of raising exception during a read
        :param strict_n: if True, by default a ValueError witll be raised if the
            exact number of requested items are not available when reading.
        """
        assert isinstance(buffer, RWLockSortedDeque)
        assert isinstance(stop_event, threading.Event)

        self._source_reader_info = source_reader_info
        self._buffer = buffer
        self._last_item = None
        self._last_key = None
        self._stop_event = stop_event
        self._sleep_time_on_iter_none_s = 0.1
        self.read_size = read_size  # read_size used by __next__
        self.peek = peek
        self.ignore_no_item_found = ignore_no_item_found
        self.strict_n = strict_n
        # _read_kwargs will be used as the defaults of the read method as well as
        # the kwargs of the __next__ method (with ignore_no_item_found=True forced)
        self._read_kwargs = dict(
            n=self.read_size,
            peek=self.peek,
            ignore_no_item_found=self.ignore_no_item_found,
            strict_n=self.strict_n,
        )
        # The _read_kwargs_for_iter should be the same as the _read_kwargs, but with
        # forced ignore_no_item_found=True so that __iter__ always returns a value, and
        # doesn't throw an error (because iter is blocking)
        self._read_kwargs_for_next = dict(self._read_kwargs, ignore_no_item_found=True)

    def __iter__(self):
        while True:
            _next = next(self)
            if _next is not None:
                yield _next
            elif self.is_stopped:
                return None
            else:
                time.sleep(self._sleep_time_on_iter_none_s)

    def __next__(self):
        # Call read with the args fixed by init
        return self.read(**self._read_kwargs_for_next)

    def set_sleep_time_on_iter_none(self, sleep_time_s: Union[int, float] = 0.1):
        """Set the sleep time of the iter yield loop when next data item is not yet available.

        :param sleep_time_s: seconds to sleep
        """
        self._sleep_time_on_iter_none_s = sleep_time_s

    def is_same_buffer(self, other_buffer_reader):
        """Check if reader is looking at the same buffer"""
        return (
            self.__class__ == other_buffer_reader.__class__
            and self._stop_event == other_buffer_reader._stop_event
        )

    @property
    def is_stopped(self) -> bool:
        """Checks if stop event has been set by StreamBuffer.

        :return: bool
        """
        return self._stop_event.is_set()

    @property
    def source_reader_info(self) -> dict:
        """A dict with important source info set by SourceReader.

        :return: dict
        """
        return self._source_reader_info

    @property
    def last_key(self):
        """key to last seen item cursor"""
        return self._last_key

    def _getlast_item(self):
        return self._last_item

    def _setlast_item(self, item):
        self._last_key = self._buffer.key(item)
        self._last_item = item

    def _dellast_item(self):
        del self._last_item
        self._last_item = None
        del self._last_key
        self._last_key = None

    last_item = property(
        _getlast_item, _setlast_item, _dellast_item, 'last seen item cursor'
    )

    def range(
        self,
        start,
        stop,
        step=None,
        *,
        peek=False,
        ignore_no_item_found=None,
        only_new_items=False,
        start_le=False,
        stop_ge=False,
    ):
        """Enables:
        1. Get last n minutes
        2. Give me data I don't have

        :param start: starting range key of item inclusive
        :param stop: ending range key of item inclusive
        :param step:
        :param peek: if True, last_item cursor will not be updated
        :param ignore_no_item_found: if True, return None when no next item
            instead of raising exception.
        :param only_new_items: if True and no new items, raise ValueError
            or return None if ignore_no_item_found
        :param start_le: if True, increase the search range to find start less than or equal by
            rounding down if start is in between keys,
            i.e. keys=[0, 10, 20], start=9 will include key=0
        :param stop_ge: if True, raise ValueError when there is no key greater than or equal to
            stop in buffer, if ignore_no_item_found is also True, return None instead of ValueError
        :return: list of items in range
        """
        if ignore_no_item_found is None:
            ignore_no_item_found = self._read_kwargs.get('ignore_no_item_found', False)
        with self._buffer.reader_lock() as reader:
            if only_new_items and self.last_key is not None:
                _next = self.read(
                    n=1, peek=True, ignore_no_item_found=ignore_no_item_found
                )
                try:
                    _next_key = reader.key(_next)
                except TypeError as e:  # TypeError: 'NoneType' object is not subscriptable
                    if ignore_no_item_found:
                        return None
                    raise e
                _start = start if start > _next_key else _next_key
            else:
                _start = start
            if start_le is True:
                with suppress(
                    ValueError
                ):  # ValueError: No item found with key at or below: _start
                    _start = reader.key(reader.find_le(_start))
            if stop_ge is True:
                try:
                    stop = reader.key(reader.find_ge(stop))
                except ValueError as e:  # ValueError: No item found with key at or above: stop
                    if ignore_no_item_found:
                        return None
                    raise e

            items = reader.range(_start, stop, step)

        if not peek:
            try:
                self.last_item = items[-1]
            except IndexError as e:  # IndexError: list index out of range
                if ignore_no_item_found:
                    return None
                raise e
        return items

    def tail(self, *, peek=False, ignore_no_item_found=False, only_new_items=False):
        """Finds the last item in buffer. Raise ValueError if no item found.

        :param peek: if True, last_item cursor will not be updated
        :param ignore_no_item_found: if True, return None when no next item
            instead of raising exception
        :param only_new_items: if True and no new items, raise ValueError
            or return None if ignore_no_item_found
        :return: tail item
        """
        with self._buffer.reader_lock() as reader:
            if only_new_items:
                try:
                    item = reader.find_last_gt(self.last_key)
                except ValueError as e:  # ValueError: No item found with key above: self.last_key
                    if ignore_no_item_found:
                        return None
                    raise e
                except TypeError as e:
                    # TypeError: '>' not supported between instances of type(key) and 'NoneType'
                    if self.last_item is None:  # first time reading a value from buffer
                        item = reader[-1]
                    else:
                        raise e
            else:
                try:
                    item = reader[-1]
                except IndexError as e:  # IndexError: deque index out of range
                    if ignore_no_item_found:
                        return None
                    raise e
        if not peek:
            self.last_item = item
        return item

    def head(self, *, peek=False):
        with self._buffer.reader_lock() as reader:
            item = reader[0]
        if not peek:
            self.last_item = item
        return item

    def read(self, n=None, *, peek=None, ignore_no_item_found=None, strict_n=None):
        """Finds an item with a key greater than the last returned item.
        Raise ValueError if no item found with key above last item.

        Note: Arguments whose values are not specified will be replaced with
        self.

        :param n: number of items to return
        :param peek: if True, last_item cursor will not be updated
        :param ignore_no_item_found: if True, return None when no next item
            instead of raising exception
        :param strict_n: if True, raise ValueError if n items are not available
        :return: next item or list of next items if n > 1
        """
        n, ignore_no_item_found, strict_n = defaulted_values(
            dict(n=n, ignore_no_item_found=ignore_no_item_found, strict_n=strict_n,),
            defaults=self._read_kwargs,
        )
        # # Is the following alternative clearer or more efficient?
        # n = n or self._read_kwargs['n']
        # peek = peek or self._read_kwargs['peek']
        # ignore_no_item_found = (
        #         ignore_no_item_found or self._read_kwargs['ignore_no_item_found']
        # )
        # strict_n = strict_n or self._read_kwargs['strict_n']

        with self._buffer.reader_lock() as reader:
            try:
                next_item = reader.find_gt(self.last_key)
            except ValueError as e:  # ValueError: No item found with key above:
                # self.last_key
                if ignore_no_item_found:
                    return None
                raise e
            except TypeError as e:
                # TypeError: '<' not supported between instances of 'NoneType' and type(key)
                if self.last_item is None:  # first time reading a value from buffer
                    next_item = reader[0]
                else:
                    raise e
            if n > 1:
                i = reader.index(next_item)
                j = i + n
                if strict_n and j >= len(reader):
                    raise ValueError(
                        f'Number of items found is less than n: strict_n={strict_n}, n={n}'
                    )

                next_items_list = reader.range_by_index(i, j)
                if not peek:
                    self.last_item = next_items_list[-1]
                return next_items_list
            if not peek:
                self.last_item = next_item
            return next_item

    def next(self, n=1, *, peek=False, ignore_no_item_found=False, strict_n=False):
        from warnings import warn

        warn(
            f'Deprecated. next is deprecated. Use read method instead',
            DeprecationWarning,
        )
        return self.read(
            n=n, peek=peek, ignore_no_item_found=ignore_no_item_found, strict_n=strict_n
        )



"""Utils for testing"""
import operator
import random
import time

from typing import Iterator
from itertools import tee
import itertools
from functools import wraps

from stream2py import SourceReader
from stream2py.utility.typing_hints import ComparableType


def consume_iterator_and_return_last_element(iterator, default=None):
    x = default
    for x in iterator:
        pass
    return x  # return last element


@wraps(itertools.islice)
def get_slice(*args, **kwargs):
    return list(itertools.islice(*args, **kwargs))


def pairwise(iterable):
    """
    >>> list(pairwise('ABCD'))
    [('A', 'B'), ('B', 'C'), ('C', 'D')]
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


# ---------------------------------------------------------------------------------------
# Online (streaming) testing of monotonicity
def is_strictly_increasing(iterable):
    return all(x < y for x, y in pairwise(iterable))


def is_strictly_decreasing(iterable):
    return all(x > y for x, y in pairwise(iterable))


def is_non_increasing(iterable):
    """
    >>> is_non_increasing([9, 2, 2, 1])
    True
    >>> is_non_increasing([9, 2, 2, 1, 6])
    False
    >>> is_non_increasing([2, 2, 2])
    True
    >>> is_non_increasing([])
    True
    """
    return all(x >= y for x, y in pairwise(iterable))


def is_non_decreasing(iterable):
    """
    >>> is_non_decreasing([1, 2, 2, 9])
    True
    >>> is_non_decreasing([1, 2, 2, 9, 6])
    False
    >>> is_non_decreasing([2, 2, 2])
    True
    >>> is_non_decreasing([])
    True
    """
    return all(x <= y for x, y in pairwise(iterable))


# More intuitive aliases
is_increasing = is_non_decreasing
is_decreasing = is_non_increasing
no_more_data = type('NoMore_data', (), {})()


def is_monotonic(iterable):
    """True if and only if iterable is non-increasing or non-decreasing

    >>> is_monotonic([1, 2, 2, 9])
    True
    >>> is_monotonic([1, 2, 2, 9, 6])
    False
    >>> is_monotonic([9, 2, 2, 1])
    True
    >>> is_monotonic([9, 2, 2, 1, 6])
    False
    >>> is_monotonic([2, 2, 2])
    True
    >>> is_monotonic([])
    True
    """
    pairs = pairwise(iterable)
    # TODO: Might be able to do something cleaner than while loop with itertools?
    while True:
        # The (no_more_data, no_more_data) is to signal end of iterator
        x, y = next(pairs, (no_more_data, no_more_data))
        if x is no_more_data:
            # Reaching here means that iterator was empty or all elements were equal
            # In any case, the iterable is monotonic, so:
            return True
        elif x != y:
            break  # you found your first non-trivial case, now the work begins

    if x < y:  # if the first (different) pairs were increasing
        # all the others must be non-decreasing
        return all(x <= y for x, y in pairs)
    else:  # if the first (different) pairs were decreasing]
        # all the others must be non-increasing
        return all(x >= y for x, y in pairs)


# ---------------------------------------------------------------------------------------
# A few source readers
from typing import Iterator, Any


class SimpleSourceReader(SourceReader):
    """A source reader that only requires an iterable"""

    _current_index = 0

    def __init__(self, data):
        """Instantiates the reader and creates an iterator if necessary"""
        if not isinstance(data, Iterator):
            data = iter(data)
        self._data = data

    def __iter__(self):
        while True:
            yield self.read()

    def open(self) -> None:
        return None

    # separate indexing concern (could be counter, could be time.time())
    def read(self):
        output = self._current_index, next(self._data)
        self._current_index += 1
        return output

    def close(self):
        return None

    @property
    def info(self) -> dict:
        return {}

    def key(self, data: Any) -> ComparableType:
        return data[0]

    def val(self, data: Any):
        return data[1]


class RandomFloatSource(SourceReader):
    """
    A simple example of source reader based on a random float generator
    """

    def __init__(self, seed=1):
        random.seed(seed)
        self.seed = seed
        self.open_count = 0

    def open(self):
        self.open_count += 1
        self.random_gen = iter(random.random, 2)

    def read(self):
        value = next(self.random_gen)
        return value

    def close(self):
        del self.random_gen

    @property
    def info(self):
        return dict(seed=self.seed, open_count=self.open_count)

    def key(self, data):
        return data


class TenthSecondCounter(SourceReader):
    """
    Example SourceReader
    Start counting when as soon as you construct
    """

    def __init__(self, starting_count=0):
        self._init_kwargs = {
            k: v for k, v in locals().items() if k not in ('self', '__class__')
        }
        self.starting_count = starting_count
        self._count = 0
        self._start_time = 0
        self.open_count = 0

    def open(self):
        """Reset params for first read"""
        self.open_count += 1
        self._count = self.starting_count
        self._start_time = int(time.time() * 10)

    @property
    def info(self):
        _info = self._init_kwargs.copy()
        _info.update(name=self.__class__.__name__, bt=self._start_time * 100000)
        return _info

    def close(self):
        """Not needed but satisfies the abstract"""

    def read(self):
        next_count = self._count
        next_time = self._start_time + self._count
        now = time.time() * 10
        if now >= next_time:
            self._count += 1
            return next_time * 100000, next_count  # (timestamp_us, count)
        return None

    key = operator.itemgetter(0)  # (timestamp_us, count) -> timestamp_us



"""Util functions. Functions that are used by, or can be used with, stream2py objects
(but do not depend on stream2py objects themselves)
"""

from inspect import signature, Parameter
from abc import abstractmethod
from functools import wraps


class TypeValidationError(TypeError):
    """Raised if an object is not valid"""

    @staticmethod
    @abstractmethod
    def is_valid(obj) -> bool:
        """Returns True if and only if obj is considered valid"""

    @classmethod
    def validate(cls, obj, *args, **kwargs):
        if not cls.is_valid(obj):
            raise cls(*args, **kwargs)


def null_enter_func(obj):
    return obj


def null_exit_func(obj, *exception_info):
    return None


def _has_min_num_of_arguments(func, mininum_n_args):
    return len(signature(func).parameters) >= mininum_n_args


def _callable_has_a_variadic_pos(func):
    """True if and only if the function has a variadic positional argument (i.e.
    *args)"""
    return any(
        x.kind == Parameter.VAR_POSITIONAL for x in signature(func).parameters.values()
    )


class EnterFunctionValidation(TypeValidationError):
    """Raised if an function isn't a valid (context manager) __enter__ function"""

    @staticmethod
    def is_valid(obj):
        return callable(obj) and _has_min_num_of_arguments(obj, 1)


class ExitFunctionValidation(TypeValidationError):
    """Raised if an function isn't a valid (context manager) __exit__ function"""

    @staticmethod
    def is_valid(obj):
        return callable(obj) and (
            _callable_has_a_variadic_pos or _has_min_num_of_arguments(obj, 4)
        )


# TODO: If needed:
#  A more general Contextualizer would proxy/delegate all methods (including special
#  ones) to the wrapped
class ContextualizeCall:
    """
    Wraps a callable so that it also becomes a context manager.

    Both enter and exit functions can be specified.

    This can be used, for instance, when the proper use of a function depends on some
    objects being "in context" (i.e. calling the enter/exit of some objects).

    This happens, for instance, when we want to use a method that depends on its
    instance being in context. For convenience, this use case is implemented by the
    ``contextualize_with_instance`` function of this module.

    Below is a simple doctest, but check out more significant tests in the tests folder.
    You can also check out:

    This issue: https://github.com/i2mint/stream2py/issues/18

    This wiki: https://github.com/i2mint/stream2py/wiki/Forwarding-context-management

    >>> f = lambda x: x + 1
    >>> hasattr(f, '__enter__') and hasattr(f, '__exit__')
    False
    >>> g = ContextualizeCall(f)
    >>> hasattr(g, '__enter__') and hasattr(g, '__exit__')
    True
    >>> with g:
    ...     print(g(2))
    3

    In the example above we get the default null enter and exit functions.
    They do nothing.
    Let's specify some that do something:

    >>> h = ContextualizeCall(f,
    ...     enter_func=lambda self: print('entering...'),
    ...     exit_func=lambda self, *errs: print('exiting...'),
    ... )
    >>> with h:
    ...     print(h(2))
    entering...
    3
    exiting...
    """

    def __init__(self, func, enter_func=null_enter_func, exit_func=null_exit_func):
        # validation
        if not callable(func):
            raise TypeError(f'First argument should be a callable, was: {func}')
        EnterFunctionValidation.validate(
            enter_func,
            f'Not a valid enter function (should be a callable with at least one arg): '
            f'{enter_func}',
        )
        ExitFunctionValidation.validate(
            exit_func,
            f'Not a valid exit function (should be a callable with at least one arg or '
            f'varadic args): {exit_func}',
        )
        # assignment
        self.func = func
        self.enter_func = enter_func
        self.exit_func = exit_func
        # wrapping self with attributes of func (signature, name, etc.)
        wraps(func)(self)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __enter__(self):
        #         print(f'{type(self).__name__}.__enter__')
        return self.enter_func(self.func)

    def __exit__(self, *exception_info):
        #         print(f'{type(self).__name__}.__exit__')
        return self.exit_func(self.func, *exception_info)


# ---------------------------------------------------------------------------------------
# Using these general objects for the particular case of having bound methods forward
# their context
# management to the instances they're bound to

from functools import partial


def forward_to_instance_enter(obj):
    return obj.__self__.__enter__()


def forward_to_instance_exit(obj, *exception_info):
    return obj.__self__.__exit__(obj, *exception_info)


contextualize_with_instance = partial(
    ContextualizeCall,
    enter_func=forward_to_instance_enter,
    exit_func=forward_to_instance_exit,
)
contextualize_with_instance.__doc__ = (
    'To be applied to a bound method. '
    'Returns a callable that forwards context enters/exits to the bound instance. '
    'See ``tests.test_util.test_contextualize_with_instance`` for and example use.'
)



from stream2py.utility.locked_sorted_deque import RWLockSortedDeque
import pytest
import operator


@pytest.mark.parametrize('initial_size,maxlen', [(100, 110), (10, 12),])
def test_rw_lock_sorted_deque(initial_size, maxlen):
    locked_deque = RWLockSortedDeque(
        ((('plc', i), f'data_{i}') for i in range(initial_size)),
        key=operator.itemgetter(0),
        maxlen=maxlen,
    )
    assert len(locked_deque) == initial_size

    new_item = (('plc', initial_size + 1), 'new data')
    # add an item and check that the len incremented accordingly
    with locked_deque.writer_lock() as writer:
        writer.append(new_item)
    assert len(locked_deque) == initial_size + 1

    # looking for the first element greater than all of those in the initial deque
    with locked_deque.reader_lock() as reader:
        assert reader.find_gt(('plc', initial_size)) == new_item

    # the deque cannot exceed its maxlen
    with locked_deque.writer_lock() as writer:
        for i in range(maxlen):
            new_item = (('plc', initial_size + 2 + i), 'new data')
            writer.append(new_item)
        assert len(locked_deque) == maxlen

    # attempting to add another element, not larger than the previous one, will raise an error
    with locked_deque.writer_lock() as writer:
        with pytest.raises(ValueError):
            writer.append(new_item)



"""
HTTP Response streaming reader with no external dependencies
--------------
.. autoclass:: stream2py.sources.net.HTTPResponseReader()
    :members:
    :show-inheritance:

    .. automethod:: __init__

"""

import urllib.request
from typing import Union, Optional
import time

from stream2py import SourceReader
from stream2py.utility.typing_hints import ComparableType


class HTTPResponseReader(SourceReader):
    _request: urllib.request.Request
    _n_bytes_read: int
    _init_kwargs: dict
    _url: str
    _response = ''

    def __init__(
        self,
        url,
        *,
        method: str = 'GET',
        headers: Optional[dict] = None,
        body: Union[str, bytes] = '',
        response_format: str = 'bytes',
        response_read_args: tuple = (),
        encoding: str = '',
    ):
        """
        :param method: The HTTP method to use.
        :param headers: The HTTP headers to pass with the request.
        :param body: The body to send with the request. Must be a string or bytes (JSON bodies should be passed through json.dumps)
        :param response_format: Format to map the response to. Accepts 'bytes' or 'str' (defaults to 'bytes')
        """
        headers = headers or {}
        self._init_kwargs = dict(
            method=method,
            headers=headers,
            body=body,
            response_format=response_format,
            encoding=encoding,
        )
        self.data = None
        if method.upper() == 'GET':
            body = None

        self._url = url
        self._n_bytes_read = 0
        self._request = urllib.request.Request(
            url, data=body, headers=headers, method=method
        )
        if response_read_args is None:
            response_read_args = ()
        if not isinstance(response_read_args, tuple):
            response_read_args = (response_read_args,)
        self.response_read_args = response_read_args

    @property
    def info(self) -> dict:
        """
        Provides a dict with init kwargs and read status

        >>> from pprint import pprint
        >>> source = HTTPResponseReader('http://zombo.com')
        >>> pprint(source.info) # doctest: +SKIP
        {'url': 'http://zombo.com',
         'n_bytes_read': 0,
         'init_kwargs': {'method': 'GET,
                         'headers': {},
                         'body': None,
                         'response_format': 'bytes'}}

        :return: dict
        """
        return {
            'url': self._url,
            'n_bytes_read': self._n_bytes_read,
            'init_kwargs': self._init_kwargs,
        }

    def key(self, data) -> ComparableType:
        """
        Returns a slice of the data retrieved.

        :param data: (start_index, end_index)
        :return: Union[str, bytes]
        """
        return data
        # return self._data[data[0]:data[1]]

    def open(self):
        request = urllib.request.Request(
            self._url, headers=self._init_kwargs['headers']
        )
        self._response = urllib.request.urlopen(request)

    def close(self):
        if self._response:
            self._response.__exit__()
            self._response = None

    def read(self):
        """Returns a chunk of raw binary data.
        """
        encoding = self._init_kwargs.get('encoding')
        result = self._response.read(*self.response_read_args)
        timestamp = time.time()
        if encoding:
            result = result.decode(encoding)
        return timestamp, result


def _test_run_HTTPResponseReader(url, **kwargs):
    """Run HTTPResponseReader and print some stuff

    :param readerClass: HTTPResponseReader class or subclass
    """
    from pprint import pprint

    output = ''

    source = HTTPResponseReader(url, **kwargs)
    pprint(source.info)
    try:
        source.open()
        pprint(source.info)
        i = 0
        while i < 600:
            data = source.read(10)
            if not data:
                break
            i += 1
            output += data

    finally:
        source.close()
        pprint(source.info)
        pprint(output)


if __name__ == '__main__':
    _test_run_HTTPResponseReader('https://zombo.com', encoding='utf-8')



from stream2py.tests.utils_for_testing import TenthSecondCounter
from stream2py.stream_buffer import StreamBuffer
import time


def test_stream_buffer():

    sc = TenthSecondCounter()  # source
    sc_buf = StreamBuffer(source_reader=sc, maxlen=100)  # buffer

    # start source, print some data, and check last_item cursor is working
    sc_buf.start()  # start reader to buffer thread
    sc_reader = (
        sc_buf.mk_reader()
    )  # reader must be made after start to have data related to said start
    time.sleep(2)
    assert sc_reader.source_reader_info['bt'] != 0, sc_reader.source_reader_info
    print('data as shown is (timestamp, count) at every tenth of a second')
    while True:
        x = sc_reader.read(n=1, ignore_no_item_found=True)
        if x:
            last_read = x
            print(x)
        else:
            break
    assert sc_reader.last_item == last_read, 'last_item did not follow last next() call'

    # check range works and last_item cursor is working
    rstart = 5
    rstop = 10
    source_info = sc_reader.source_reader_info
    start_key = source_info['bt'] + 1e5 * rstart  # 5
    stop_key = source_info['bt'] + 1e5 * rstop  # 10
    range_data = sc_reader.range(start_key, stop_key)
    expected_data = None
    for expected_data, rdata in zip(range(rstart, rstop + 1), range_data):
        timestamp, _data = rdata
        assert _data == expected_data, 'wrong data does not match'
    assert (
        expected_data == sc_reader.last_item[1]
    ), f'last_item did not follow last range({start_key}, {stop_key}) call'
    previous_last_item = expected_data

    # check range works with step and peek, and last_item does not move with peek=True
    rstart = 10
    rstop = 15
    rstep = 2
    source_info = sc_reader.source_reader_info
    start_key = source_info['bt'] + 1e5 * rstart  # 5
    stop_key = source_info['bt'] + 1e5 * rstop  # 10
    range_data = sc_reader.range(start_key, stop_key, step=rstep, peek=True)
    for expected_data_value, rdata in zip(range(rstart, rstop + 1, rstep), range_data):
        timestamp, _data = rdata
        assert _data == expected_data_value, 'wrong data does not match'
    assert (
        previous_last_item == sc_reader.last_item[1]
    ), 'last_item moved but should not when peek=True'

    # stop source and check if reader see it
    sc_reader12 = sc_buf.mk_reader()  # reader1
    assert sc_reader.is_same_buffer(sc_reader12), 'first readers should be equal'
    assert (
        sc_reader.last_item != sc_reader12.last_item
    ), 'first readers should have a different last_item cursor position'
    assert sc_reader.is_stopped is False, 'Reader should see source is not stopped'
    sc_buf.stop()
    assert sc_reader.is_stopped is True, 'Reader should see source is stopped'

    # restart source and make new readers then check old reader is different from new reader
    with sc_buf:  # start and stop with a context manager

        sc_reader21 = sc_buf.mk_reader()  # reader2
        sc_reader22 = sc_buf.mk_reader()  # reader2
        assert sc_reader21.is_same_buffer(
            sc_reader22
        ), 'two new readers should be equal'
        assert not sc_reader.is_same_buffer(
            sc_reader21
        ), 'first reader should not equal new reader'
        assert (
            sc_reader.is_stopped is True
        ), 'first reader should still see a stopped source'
        assert (
            sc_reader21.is_stopped is False
        ), 'new reader should see the source is running'
        assert (
            sc_reader22.last_item == sc_reader21.last_item
        ), 'new readers should now have the same last_item cursor position'

        # print some data from old and new reader and see they are different
        time.sleep(1.5)
        print('comparing first start reader values to second start')
        while True:
            r22 = sc_reader22.read(n=1, ignore_no_item_found=True)
            r12 = sc_reader12.read(n=1, ignore_no_item_found=True)

            if r22:

                def zero_if_none(val):
                    if val is None:
                        return 0, 0
                    return val

                print(f'{r22} > {r12}, ts_diff={r22[0] - r12[0]}')
                assert r22 > zero_if_none(r12)
            else:
                break
        assert (
            sc_reader22.last_item != sc_reader21.last_item
        ), 'new readers should now have a different last_item cursor position'



"""
A StreamBuffer has 2 jobs: First, it manages the open, read, and close of a SourceReader
and puts read data onto a thread-safe buffer.
Second, it is a factory of BufferReaders instances for multiple consumers.
"""
__all__ = ['StreamBuffer']

import logging
import threading
import time
from typing import Optional, Union

from stream2py import BufferReader, SourceReader
from stream2py.utility.locked_sorted_deque import RWLockSortedDeque

logger = logging.getLogger(__name__)

DFLT_SLEEP_TIME_ON_READ_NONE_S = 0.3


class _SourceBuffer:
    def __init__(
        self,
        source_reader_info: dict,
        stop_event: threading.Event,
        *,
        key=None,
        maxlen: int = 10000,
    ):
        """StreamBuffer helper class
        Thread safe buffer for reading and writing data items

        source_name should default to source class name
        source_info should default to args for constructing source class
        an audio source info should have info about sample rates, bit depth, etc.
        reader should get both source_name and source_info
        """
        self._stop_event = stop_event
        self._buffer = RWLockSortedDeque([], key=key, maxlen=maxlen)
        self._source_reader_info = source_reader_info

    def __len__(self):
        return len(self._buffer)

    @property
    def key(self):
        return self._buffer.key

    @property
    def source_reader_info(self):
        return self._source_reader_info

    def append(self, item):
        with self._buffer.writer_lock() as writer:
            writer.append(item)

    def drop(self, n=1):
        with self._buffer.writer_lock() as writer:
            writer.drop(n)

    def mk_reader(self):
        return BufferReader(
            buffer=self._buffer,
            source_reader_info=self._source_reader_info,
            stop_event=self._stop_event,
        )


class StreamBuffer:
    """Handles starting and stopping SourceReader and making BufferReaders

    >>> from stream2py import StreamBuffer
    >>> from stream2py.examples.source_reader import SimpleCounterString
    >>>
    >>> source_reader = SimpleCounterString(start=0, stop=100)
    >>> stream_buffer = StreamBuffer(source_reader=source_reader, maxlen=100)
    >>> stream_buffer.is_running
    False
    >>> print(stream_buffer.source_reader_info)
    None
    >>> stream_buffer.start()
    >>> assert stream_buffer.is_running
    >>> stream_buffer.source_reader_info
    {'start': 0, 'stop': 100, 'open_count': 1}
    >>> open1_reader1 = stream_buffer.mk_reader()
    >>> open1_reader1.next()
    's0'
    >>> open1_reader2 = stream_buffer.mk_reader()
    >>> # readers from the same open instance are the same
    >>> assert open1_reader1.is_same_buffer(open1_reader2) is True
    >>> assert open1_reader1.next() != open1_reader2.next()  # item cursor position is different
    >>> stream_buffer.stop()
    >>>
    >>> with stream_buffer:
    ...     stream_buffer.source_reader_info
    ...     open2_reader1  = stream_buffer.mk_reader()
    ...     open2_reader2  = stream_buffer.mk_reader()
    ...     # readers from the same open instance are the same
    ...     assert open2_reader1.is_same_buffer(open2_reader2) is True
    ...     # readers from the different open instances
    ...     assert open2_reader1.is_same_buffer(open1_reader1) is False
    {'start': 0, 'stop': 100, 'open_count': 2}
    """

    def __init__(
        self,
        source_reader: SourceReader,
        *,
        maxlen: int,
        sleep_time_on_read_none_s: Optional[Union[int, float]] = None,
        auto_drop=True,
    ):
        """
        TODO: option to auto restart source on read exception

        :param source_reader: instance of a SourceReader subclass
        :param maxlen: max number of read data points to store in buffer before data starts dropping
            off the queue
        :param sleep_time_on_read_none_s: Seconds to sleep when reading None from source_reader.
            None to use defaults.
        :param auto_drop: False to stop reading when buffer is full and use StreamBuffer.drop() to
            manually make space.
        """
        assert isinstance(
            source_reader, SourceReader
        ), 'source_reader is not a subclass of SourceReader'
        self.source_reader = source_reader
        self._maxlen = maxlen
        self.auto_drop = auto_drop
        if isinstance(sleep_time_on_read_none_s, (int, float)):
            self._sleep_time_on_read_none_s = sleep_time_on_read_none_s
        elif isinstance(source_reader.sleep_time_on_read_none_s, (int, float)):
            self._sleep_time_on_read_none_s = source_reader.sleep_time_on_read_none_s
        else:
            self._sleep_time_on_read_none_s = DFLT_SLEEP_TIME_ON_READ_NONE_S  # default

        self._stop_event = None
        self.source_buffer = None
        # start_lock used to lock mk_reader while source is still starting up
        self.start_lock = threading.Lock()
        self._next_reader = None
        self._read_to_buffer_thread = None

    def __iter__(self):
        reader = self.mk_reader()
        yield from iter(reader)

    def __next__(self):
        if self._next_reader is None:
            self._next_reader = self.mk_reader()
        return next(self._next_reader)

    def drop(self, n=1):
        """Manually drop items from buffer when auto_drop is False

        :param n: number of items to drop from the left side
        :return:
        """
        if self.auto_drop is True:
            raise RuntimeError(
                'auto_drop must be False to manually drop items from buffer'
            )
        self.source_buffer.drop(n)

    def start(self):
        """Open and start reading from source_reader into buffer"""
        with self.start_lock:
            if self._stop_event and not self._stop_event.is_set():
                self.stop()
            self._set_read_to_buffer_thread_and_source_buffer()
            self._open()
            self._read_to_buffer_thread.start()

    def stop(self):
        """Stop reading and close source_reader"""
        self._stop_event.set()
        time.sleep(1)
        self._next_reader = None

    def mk_reader(self) -> BufferReader:
        """ Makes a BufferReader instance for the currently running  StreamBuffer.
        Reader must be made after start() to have data from said start.

        :return: BufferReader instance
        """
        with self.start_lock:
            if not isinstance(self.source_buffer, _SourceBuffer):
                raise RuntimeError('Readers should be made after starting')
            return self.source_buffer.mk_reader()

    @property
    def source_reader_info(self) -> Optional[dict]:
        """A dict with important source info set by SourceReader.

        :return: dict or None
        """
        if self.source_buffer:
            return self.source_buffer.source_reader_info
        return None

    @property
    def is_running(self) -> bool:
        """Checks if stop event has been set.

        :return: bool
        """
        with self.start_lock:
            return self._stop_event is not None and not self._stop_event.is_set()

    @property
    def join(self):
        return self._read_to_buffer_thread.join

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    enter, exit = __enter__, __exit__

    def _run(self):
        try:
            while not self._stop_event.is_set():
                # check if buffer is full and skip read or append
                if self.auto_drop is True or len(self.source_buffer) < self._maxlen:
                    data = self.source_reader.read()
                else:
                    data = None

                if data is not None:
                    self.source_buffer.append(data)
                else:
                    time.sleep(self._sleep_time_on_read_none_s)
        finally:
            if not self._stop_event.is_set():
                self._stop_event.set()
            try:
                self.source_reader.close()
            except Exception as e:
                logger.error(e)

    def _set_read_to_buffer_thread_and_source_buffer(self):
        """Initialize source and buffer
        Makes a new run thread, and stop event that is unique to this run session"""
        if self._stop_event and not self._stop_event.is_set():
            self.stop()
        self.source_buffer = None
        self._read_to_buffer_thread = threading.Thread(target=self._run, daemon=True)
        self._stop_event = threading.Event()

    def _open(self):
        """First thing called after start/run
        Calls source_reader.open() and then sets up source_buffer with latest
        source_reader.info"""
        self.source_reader.open()
        self.source_buffer = _SourceBuffer(
            source_reader_info=self.source_reader.info,
            stop_event=self._stop_event,
            key=self.source_reader.key,
            maxlen=self._maxlen,
        )

    def _mk_contextualized_iterator(self):
        """Return next item (entering the context beforehand, if not running).
        This method is meant to be called under context so that a clean exit is assured.
        """
        if not self.is_running:
            self.__enter__()
        return self.__next__()

    def __call__(self):
        """Return next item (entering the context beforehand, if not running).
        This method is meant to be called under context so that a clean exit is assured. """
        return self._mk_contextualized_iterator()

    def __del__(self):
        try:
            self.__exit__(None, None, None)
        except Exception:
            pass



"""
A SourceReader defines how to get data with the methods: open(), read(), and close(), and also how
the data is ordered with the key() method and an info property describing the instance."""
__all__ = ['SourceReader']


from abc import ABCMeta, abstractmethod
import time
from typing import Optional, Any, Union, NewType, Iterator

from stream2py.utility.typing_hints import ComparableType


class SourceReader(metaclass=ABCMeta):
    """Abstract class interface to be used by StreamBuffer.

    >>> from stream2py import SourceReader
    >>>
    >>> class SimpleCounterString(SourceReader):
    ...     '''Count in range returning a string formatted as f"s{count}"'''
    ...     def __init__(self, start, stop):
    ...         assert start < stop
    ...         self.start = start
    ...         self.stop = stop
    ...         self.range_iterator = None
    ...         self.open_count = 0
    ...
    ...     def open(self):
    ...         self.open_count += 1
    ...         self.range_iterator = iter(range(self.start, self.stop))
    ...
    ...     def read(self):
    ...         value = next(self.range_iterator, None)
    ...         if value is not None:
    ...             return f's{value}'
    ...         else:
    ...             return None
    ...
    ...     def close(self):
    ...         del self.range_iterator
    ...         self.range_iterator = None
    ...
    ...     @property
    ...     def info(self):
    ...         return dict(start=self.start, stop=self.stop, open_count=self.open_count)
    ...
    ...     def key(self, data):
    ...         return int(data[1:])
    ...
    >>> source_reader = SimpleCounterString(start=0, stop=10)
    >>> source_reader.open()
    >>> source_reader.info
    {'start': 0, 'stop': 10, 'open_count': 1}
    >>> source_reader.read()
    's0'
    >>> source_reader.close()
    >>> with source_reader:
    ...     source_reader.info
    ...     source_reader.read()
    ...     source_reader.read()
    ...     source_reader.read()
    {'start': 0, 'stop': 10, 'open_count': 2}
    's0'
    's1'
    's2'
    """

    _sleep_time_on_iter_none_s = 0.001

    def __iter__(self):
        while True:
            _next = self.read()
            if _next is not None:
                yield _next
            else:
                time.sleep(self._sleep_time_on_iter_none_s)

    def set_sleep_time_on_iter_none(self, sleep_time_s: Union[int, float] = 0.001):
        """Set the sleep time of the iter yield loop when next data item is not yet available.

        :param sleep_time_s: seconds to sleep
        """
        self._sleep_time_on_iter_none_s = sleep_time_s

    @abstractmethod
    def open(self) -> None:
        """Set up source to be read and set some source info affected by open time like the time of
        open. Will be called in StreamBuffer immediately before first read."""

    @abstractmethod
    def close(self) -> None:
        """Close and clean up source reader.
        Will be called when StreamBuffer stops or if an exception is raised during read and append
        loop.
        """

    @property
    def info(self) -> dict:
        """A dict with important source info. Default can be init_kwargs and open timestamp.

        This info will be available in BufferReader after SourceReader has opened but any changes
        made between open and close are not guaranteed to be visible. Any continuous status updates
        should be included with read data instead.

        :return: dict
        """
        return {}

    @abstractmethod
    def key(self, data: Any) -> ComparableType:
        """
        Converts data into a comparable value to sort by

        :param data: the return value of the 'read' method
        :return: ComparableType
        """
        raise NotImplementedError(
            "Implement the 'key' method to convert data into a comparable value to sort by"
        )

    @property
    def sleep_time_on_read_none_s(self) -> Optional[Union[int, float]]:
        """Sets default sleep time for StreamBuffer when it reads None from SourceReader.
        Useful when you know the period between data points.

        For example if data only comes once every 10 seconds,
        it would be inefficient to try and read again every 0.1 seconds and get None up to 100 times
        but every 10 seconds or even 1 second would be much more reasonable depending on the use
        case. Not necessary if read is a blocking function that always returns a data point.

        Default is None which delegates the sleep time to StreamBuffer's default.

        :return: Optional[Union[int, float]] number of seconds to sleep
        """
        return None

    timestamp_seconds_to_unit_conversion: Union[int, float] = 1e6  # to microseconds

    @classmethod
    def get_timestamp(cls) -> int:
        """Get timestamp of now as int microseconds

        :return: timestamp in UTC microseconds
        """
        return int(time.time() * cls.timestamp_seconds_to_unit_conversion)

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    enter, exit = __enter__, __exit__

    def stream_buffer(
        self,
        maxlen: int,
        sleep_time_on_read_none_s: Optional[Union[int, float]] = None,
        auto_drop=True,
    ):
        from stream2py.stream_buffer import StreamBuffer

        return StreamBuffer(
            self,
            maxlen=maxlen,
            sleep_time_on_read_none_s=sleep_time_on_read_none_s,
            auto_drop=auto_drop,
        )


StreamItem = NewType('StreamItem', Any)


class QuickSourceReader(SourceReader):
    """
    A `SourceReader` with whose only required merthod is the read method.

    - open just sets an `open_time` attribute and a `read_idx` attribute (to 0)
    - close does nothing
    - info returns a `{`open_time`: ..., `open_instance`: ...}` dict
    - key just returns the data itself
    """

    read_idx = None
    open_time = None

    def __iter__(self) -> Iterator[StreamItem]:
        while True:
            _next = self.read()
            if self.is_valid_data(_next):
                yield _next
            else:
                self.when_no_data(_next)

    def is_valid_data(self, data: StreamItem) -> bool:
        """Bool function indicating whether the data should be released in an iteration.

        By default, all non-None data will be considered valid, but some streams
        might indicate "empty data" in a different way.
        In that case, `is_valid_data` can be used to adapt to the situation.

        This filter function is meant to indicate "empty data", or "don't yield this
        data" during iteration, but can also be used simply to filter out data that
        we don't want to release in the iteration.
        """
        return data is not None

    def when_no_data(self, _next=None):
        """Action taken when data is not valid (not yielded in iteration).
        Usually, the action is to sleep for a fixed amount of seconds to give a
        chance for the source data production to catch up, without hogging resources
        by waiting.
        The `when_no_data` can do much more though, and even use the source_reader
        instance and/or the (non valid) data itself to decide what action to take.

        For example, one may want to have when_no_data do some timing stats to decide
        on the number of seconds to stop, or send an interrupt signal if too many
        "non datas" happen in a row.
        """
        time.sleep(self._sleep_time_on_iter_none_s)

    def open(self):
        self.read_idx = 0
        self.open_time = self.get_timestamp()

    def close(self):
        """A dummy close class"""

    @property
    def info(self):
        return {
            'open_time': self.open_time,
            'open_instance': str(self.open_instance),
        }

    def key(self, data: StreamItem):
        return data



import pytest

from stream2py.tests.utils_for_testing import RandomFloatSource, TenthSecondCounter

# note that the class to test must have the additional attribute open_count for the test to pass
@pytest.mark.parametrize(
    'n_reads,n_open_close,source_reader_class,class_params',
    [
        (1, 10, RandomFloatSource, {'seed': 1}),
        (5, 100, RandomFloatSource, {'seed': 1}),
        (10, 10, RandomFloatSource, {'seed': 1}),
        (10, 10, TenthSecondCounter, dict()),
    ],
)
def test_source_readers_open_close(
    n_reads, n_open_close, source_reader_class, class_params
):
    sc = source_reader_class(**class_params)
    for open_number in range(n_open_close):
        with sc:
            # check that the count increments
            assert sc.open_count == open_number + 1
            random_numbers = [sc.read() for i in range(n_reads)]
            # check that each read yielding one float
            assert len(random_numbers) == n_reads
        # check that after closing, the count of open did not increment
        assert sc.open_count == open_number + 1
        # checking access to the info
        assert isinstance(sc.info, dict)
